# -*- indent-tabs-mode: nil -*-

require 'monitor'

class Koya
  VERSION = '0.1'

  class KoyaError < RuntimeError; end
  class TransactionNotFound < KoyaError; end
  class CanNotRevertError < RuntimeError; end

  module DictMixin
    def [](key)
      @_koya_[key.to_s]
    end

    def []=(key, value)
      @_koya_[key.to_s] = value
    end
    
    def each(&block)
      @_koya_.each(&block)
    end

    def keys
      @_koya_.keys
    end
  end

  class KoyaObject
    def self.bless(ref)
      it = self.allocate
      it.instance_eval {
        @_koya_ = ref
      }
      it
    end

    def self.new(*arg)
      s = Thread.current['_koya_'].to_a[-1]
      raise TransactionNotFound unless s
      ref = s.create_object(self)
      it = bless(ref)
      it.instance_eval {
        initialize(*arg)
      }
      it
    end

    def self.koya_attr_reader(*arg)
      arg.each do |name|
        module_eval("def #{name}; @_koya_['#{name}']; end")
      end
    end

    def self.koya_attr_writer(*arg)
      arg.each do |name|
        module_eval("def #{name}=(v); @_koya_['#{name}']=v; end")
        module_eval("alias set_#{name} #{name}=")
      end
    end

    def self.koya_attr(*arg)
      koya_attr_reader(*arg)
      koya_attr_writer(*arg)
    end

    def transaction(*arg)
      @_koya_.transaction do |koya|
        yield(koya)
      end
    end
    
    attr_reader :_koya_

    def ==(other)
      self.class == other.class && @_koya_ == other._koya_
    end
  end

  class Dict < KoyaObject
    include DictMixin
  end

  class Stream < KoyaObject
    def initialize
      self.head = 0
      self.size = 0
    end
    koya_attr(:head)
    koya_attr(:size)

    def [](key, length=nil)
      transaction do
        head = self.head
        size = self.size

        range_or_key = to_range(key, length)

        if Range === range_or_key
          range_or_key.collect do |k|
            index = to_index(head, size, k)
            index ? fetch(index) : nil
          end
        elsif Integer === range_or_key
          index = to_index(head, size, range_or_key)
          index ? fetch(index) : nil
        else
          nil
        end
      end
    end

    def push(value)
      transaction do
        head = self.head
        size = self.size
        store(head + size, value)
        self.size = size + 1
      end
    end

    def <<(value)
      push(value)
    end

    def unshift(value)
      transaction do
        head = self.head
        size = self.size
        head -= 1
        store(head, value)
        self.head = head
        self.size = size + 1
      end
    end

    def pop
      transaction do
        head = self.head
        size = self.size

        return nil if size <= 0

        self.size = size - 1
        fetch(head + size - 1, true)
      end
    end
    
    def shift
      transaction do
        head = self.head
        size = self.size

        return nil if size <= 0

        self.head = head + 1
        self.size = size - 1

        fetch(head, true)
      end
    end
    
    def to_a
      transaction do
        head = self.head
        size = self.size

        (head...(head + size)).inject([]) do |ary, i|
          ary << fetch(i)
        end
      end
    end

    def each(&block)
      to_a.each(&block)
    end

    private
    def to_range(key, length)
      if length
        return nil if length < 0
        key...(length + key)
      elsif Range === key
        key
      else
        key
      end
    end

    def to_index(head, size, key)
      key = size + key if key < 0

      if (0...size) === key
        head + key
      else
        false
      end
    end

    def store(index, value)
      @_koya_[index] = value
    end
    
    def fetch(index, remove=false)
      @_koya_[index]
    ensure
      @_koya_.delete(index) if remove
    end
  end

  class KoyaRef
    def initialize(store, rowid)
      @store = store
      @rowid = rowid
    end
    attr_reader :rowid, :store

    def ==(other)
      self.class == other.class &&
      self.store == other.store &&
      self.rowid == other.rowid
    end

    def bless
      klass = @store.get_klass(@rowid)
      case klass
      when '@'
        self
      else
        begin
          return eval("#{klass}.bless(self)")
        rescue
          self
        end
      end
    end

    def to_s
      @rowid.to_s
    end

    def transaction
      @store.transaction do |h|
        yield(h)
      end
    end

    def [](key)
      @store.get_prop(@rowid, key)
    end

    def []=(key, value)
      @store.set_prop(@rowid, key, value)
    end

    def delete(key)
      @store.delete_prop(@rowid, key)
    end

    def keys
      @store.prop_keys(@rowid)
    end

    def each(&block)
      transaction do
        @store.all_prop(@rowid).each(&block)
      end
    end

    def to_a
      transaction do
        @store.all_prop(@rowid)
      end
    end

    def lock
      @store.lock_object(@rowid)
    end

    def unlock
      @store.unlock_object(@rowid)
    end

    def referer
      @store.referer(@rowid)
    end

    def revisions(name)
      @store.revisions(@rowid, name)
    end

    def _koya_; self; end
  end

  class Store
    include MonitorMixin

    begin
      require 'sqlite'
      Database = SQLite::Database
      SQLException = SQLite::Exceptions::SQLException
    rescue LoadError
      require 'sqlite3'
      Database = SQLite3::Database
      SQLException = SQLite3::SQLException
    end

    def initialize(name, use_log=true)
      super()
      @name = name
      @conn = Database.new(name)
      @conn.busy_handler {|a, b| on_busy(a, b) }
      @polling_interval = 0.1
      @in_transaction = nil
      @use_log = use_log
      @root = setup_root(use_log)
      setup_version
    end
    attr_reader :root, :version, :use_log
    attr_accessor :polling_interval

    def on_busy(resource, retries)
      sleep(@polling_interval)
      true
    end

    def time_to_rev(time)
      sprintf("%016.5f", time)
    end

    def transaction
      synchronize do
        return yield(@conn) if @in_transaction
        begin
          stack = Thread.current['_koya_'] ||= []
          stack.push(self)
          @in_transaction = time_to_rev(Time.now)
          @conn.transaction do |h|
            return yield(h)
          end
        ensure
          stack.pop
          @in_transaction = nil
        end
      end
    end

    def vacuum
      synchronize do
        @conn.execute('vacuum') unless @in_transaction
      end
    end

    def setup_root(use_log = true)
      begin
        create_table(use_log)
      rescue SQLException
      end
      
      create_root
    end

    def create_table(use_log=true)
      transaction do |h|
        h.execute("create table version (koya, log)")
        h.execute("insert into version (koya, log) values (?, ?)",
                  VERSION, use_log ? 1 : 0)
        h.execute("create table memory (klass, lock, gc, gen)")
        h.execute("create table prop (object, name, klass, value, time, unique(object, name))")
        h.execute("create table prop_log (object, name, klass, value, time, unique(object, name, time))")
      end
    end

    def create_root
      transaction do |h|
        rowid = h.get_first_value('select rowid from memory where klass=?',
                                  'root')
        return KoyaRef.new(self, rowid) if rowid

        h.execute("insert into memory (klass, gen) values (?, 0)", 'root')
        rowid = h.get_first_value("select last_insert_rowid();")
        return KoyaRef.new(self, rowid)
      end
    end

    def dump_object(obj)
      case obj
      when KoyaObject
        return '@', obj._koya_.to_s
      when KoyaRef
        return '@', obj.to_s
      when Fixnum
        return 'i', obj.to_i
      when String
        return 's', obj.to_s
      else
        begin
          return 'r', Marshal.dump(obj).dump
        rescue
          return 'R', obj.inspect
        end
      end
    end

    def load_object(klass, value)
      case klass
      when '@'
        KoyaRef.new(self, value).bless
      when 'i'
        value.to_i
      when 's'
        value.to_s
      when 'r'
        Marshal.load(eval(value))
      when 'R'
        value.to_s
      else
        nil
      end
    end

    def create_object(klass)
      transaction do |h|
        h.execute("insert into memory (klass, gen)  values (?, 0)", klass.to_s)
        rowid = h.get_first_value("select last_insert_rowid();")
        KoyaRef.new(self, rowid)
      end
    end

    def lock_object(rowid)
      transaction do |h|
        locked = h.get_first_value("select lock from memory where rowid=?",
                                   rowid)
        return false if locked 
        h.execute("update memory set lock=1 where rowid=?", rowid)
        return true
      end
    end

    def unlock_object(rowid)
      transaction do |h|
        h.execute("update memory set lock=null where rowid=?", rowid)
      end
    end

    def get_klass(rowid)
      transaction do |h|
        h.get_first_value("select klass from memory where rowid=?", rowid)
      end
    end

    def set_prop(rowid, name, obj)
      transaction do |h|
        k, v = dump_object(obj)
        h.execute("insert or replace into prop values (?, ?, ?, ?, ?)",
                  rowid, name.to_s, k, v, @in_transaction)
        h.execute("insert or replace into prop_log values (?, ?, ?, ?, ?)",
                  rowid, name.to_s, k, v, @in_transaction) if @use_log
      end
    end

    def get_prop(rowid, name)
      transaction do |h|
        k, v = h.get_first_row("select klass, value from prop where object=? and name=? and klass not null",
                               rowid, name.to_s)
        return load_object(k, v)
      end
    end

    def delete_prop(rowid, name)
      transaction do |h|
        h.execute("insert or replace into prop values (?, ?, null, null, ?)",
                  rowid, name.to_s, @in_transaction)
        h.execute("insert or replace into prop_log values (?, ?, null, null, ?)",
                  rowid, name.to_s, @in_transaction) if @use_log
      end
    end

    def prop_keys(rowid)
      transaction do |h|
        h.execute("select name from prop where object=? and klass not null",
                  rowid).collect {|k| k[0]}
      end
    end

    def all_prop(rowid)
      transaction do |h|
        h.execute("select name, klass, value from prop where object=? and klass not null",
                  rowid).collect do |v|
          [v[0], load_object(v[1], v[2])]
        end
      end
    end

    def extent(klass)
      name = klass.to_s
      transaction do |h|
        gc
        h.execute("select rowid from memory where klass=?", name).collect do |r|
          load_object('@', r[0])
        end
      end
    end

    def referer(rowid)
      transaction do |h|
        gc
        h.execute("select object, name from prop where klass=? and value=?", 
                  '@', rowid).collect do |r|
          [load_object('@', r[0]), r[1]]
        end
      end
    end

    def setup_version
      transaction do |h|
        @version, log = h.get_first_row("select koya, log from version")
        @use_log = (log.to_i > 0)
      end
    end

    def revert_to(time)
      raise CanNotRevertError unless @use_log
      rev = time_to_rev(time)
      transaction do |h|
        ary = h.execute("select object, name from prop where time>=?", rev)
        ary.each do |r|
          delete_prop(r[0], r[1])
          k, v = h.get_first_row("select klass, value from prop_log where time<=? and object=? and name=? order by time desc limit 1", rev, r[0], r[1])
          if k
            set_prop(r[0], r[1], load_object(k, v))
          end
        end
      end
    end

    def revisions(rowid=nil, name=nil)
      if rowid
        transaction do |h|
          ary = h.execute("select klass, value, time from prop_log where object=? and name=? order by time desc", rowid.to_s, name.to_s)
          ary.collect do |k, v, t|
            [Time.at(t.to_f), load_object(k, v)]
          end
        end
      else
        transaction do |h|
          h.execute("select distinct time from prop_log order by time desc").collect do |r|
            Time.at(r[0].to_f)
          end
        end
      end
    end

    def [](name)
      @root[name]
    end

    def []=(name, obj)
      @root[name] = obj
    end

    def gc
      GC.new(self).gc
    end

    def baby_gc
      BabyGC.new(self).gc
    end
  end
  
  class GC
    def initialize(store)
      @store = store
    end

    def gc
      @store.transaction do |h|
        setup_mark
        level = 1
        while walk(level)
          level += 1
        end
        delete_unmarked
        delete_mark
        update_gen
      end
      @store.vacuum
    end

    private
    def setup_mark
      @store.transaction do |h|
        h.execute("update memory set gc=null")
        h.execute("update memory set gc=0 where lock not null or rowid=?",
                  @store.root)
      end
    end

    def walk(level)
      sql_mark = <<EOQ
update memory set gc=? where gc is null and rowid in (
  select value from prop where klass='@' and object in (
    select rowid from memory where gc not null
  )
);
EOQ
      @store.transaction do |h|
        h.execute(sql_mark, level)
        count = h.get_first_value("select count(*) from memory where gc=?",
                                  level).to_i 
        return count > 0
      end
    end

    def delete_unmarked
      sql_delete_object = <<EOQ
delete from memory where gc is null;
EOQ
      sql_delete_prop = <<EOQ
delete from prop where object not in (
  select rowid from memory where gc not null
);
EOQ
      @store.transaction do |h|
        h.execute(sql_delete_prop)
        h.execute(sql_delete_object)
      end
    end

    def delete_mark
      @store.transaction do |h|
        h.execute("update memory set gc=null")
        h.execute("delete from prop_log")
      end
    end

    def update_gen
      @store.transaction do |h|
        h.execute("update memory set gen=gen+1")
      end
    end
  end

  class BabyGC < GC
    def setup_mark
      super
      @store.transaction do |h|
        h.execute("update memory set gc=0 where gen > 0")
      end
    end

    def update_gen
    end
  end
  
  def self.root(dbname)
    Store.new(dbname).root
  end
end
