require 'test/unit'
require 'koya'

class KoyaTest < Test::Unit::TestCase
  def initialize(*arg)
    super(*arg)
    @dbname = 'koya_test.db'
  end

  def setup
    @koya = Koya::Store.new(@dbname)
  end

  def teardown
    File::unlink(@dbname)
  end

  def test_create_twice
    assert_equal('0.1', @koya.version)

    first = @koya.root
    second = Koya::Store.new(@dbname).root

    assert_equal('0.1', second.store.version)

    assert_equal(first.to_s, second.to_s)
    assert_equal(nil, first['foo'])
    first['foo'] = 123.0

    assert_equal(123.0, first['foo'])
    assert_equal(123.0, second['foo'])
  end

  def test_marshal_with_zero
    @koya['array'] = [0, 1, 2, 3]

    assert_equal([0, 1, 2, 3], @koya['array'])
  end

  class MyNode < Koya::KoyaObject
    koya_attr(:left, :right, :value)

    def lock
      @_koya_.lock
    end

    def unlock
      @_koya_.unlock
    end
  end

  def test_gc
    root = @koya.root
    
    root.transaction do
      root['top'] = MyNode.new
      root['top'].left = MyNode.new
      root['top'].right = MyNode.new
    end

    top = root['top']

    right = top.right
    assert_equal(true, right.lock)
    right.value = 'memo'

    left = top.left

    left.transaction do
      left.value = Koya::Stream.new
      100.times do |n|
        left.value.push(n)
      end
    end

    assert_equal(50, left.value[50])
    root.delete('top')

    @koya.gc
    
    assert_equal(nil, left.value)
    assert_equal('memo', right.value)

    right.unlock

    @koya.gc

    assert_equal(nil, right.value)
  end
  
  class MyKoya < Koya::KoyaObject
    koya_attr('name')
  end

  def test_object
    assert_raise(Koya::TransactionNotFound) do
      MyKoya.new
    end
    @koya.transaction do
      @koya.root['tmp'] = MyKoya.new
    end

    m = @koya.root['tmp']
    assert_kind_of(MyKoya, m)
    m.name = 'foo'
    assert_equal('foo', m.name)
  end

  class MyCell < Koya::KoyaObject
    def initialize(value)
      self.value = value
    end

    koya_attr('left')
    koya_attr('right')
    koya_attr('value')
    def push(value)
      transaction do
        raise '4' if value == 4
        cell = self.class.new(value)
        raise '5' if value == 5
        cell.right = self.right
        raise '6' if value == 6
        cell.left = self
        raise '7' if value == 7
        self.right = cell
        raise '8' if value == 8
      end
    end
  end

  def test_rollback
    @koya.transaction do
      @koya.root['cell'] = MyCell.new(1)
    end
    cell = @koya.root['cell']
    cell.push(2)
    cell.push(3)

    assert_equal(1, cell.value)
    assert_equal(3, cell.right.value)
    assert_equal(2, cell.right.right.value)
    assert_equal(nil, cell.right.right.right)

    cell.push(4) rescue nil
    @koya.gc

    assert_equal(1, cell.value)
    assert_equal(3, cell.right.value)
    assert_equal(2, cell.right.right.value)
    assert_equal(nil, cell.right.right.right)

    cell.push(5) rescue nil
    @koya.gc

    assert_equal(1, cell.value)
    assert_equal(3, cell.right.value)
    assert_equal(2, cell.right.right.value)
    assert_equal(nil, cell.right.right.right)

    cell.push(6) rescue nil
    @koya.gc

    assert_equal(1, cell.value)
    assert_equal(3, cell.right.value)
    assert_equal(2, cell.right.right.value)
    assert_equal(nil, cell.right.right.right)

    cell.push(7) rescue nil
    @koya.gc

    assert_equal(1, cell.value)
    assert_equal(3, cell.right.value)
    assert_equal(2, cell.right.right.value)
    assert_equal(nil, cell.right.right.right)

    cell.push(8) rescue nil
    @koya.gc

    assert_equal(1, cell.value)
    assert_equal(3, cell.right.value)
    assert_equal(2, cell.right.right.value)
    assert_equal(nil, cell.right.right.right)

    cell.push(9)
    @koya.gc

    assert_equal(1, cell.value)
    assert_equal(9, cell.right.value)
    assert_equal(3, cell.right.right.value)
    assert_equal(2, cell.right.right.right.value)
    assert_equal(nil, cell.right.right.right.right)
  end

  def test_stream
    @koya.transaction do
      @koya.root['stream'] = Koya::Stream.new
    end
    s = @koya.root['stream']

    assert_equal(nil, s.pop)

    s.push(1)
    s.push(2)
    s.push(3)

    assert_equal(3, s[2])
    assert_equal(2, s[1])
    assert_equal(1, s[0])
    assert_equal(3, s[-1])
    assert_equal(2, s[-2])
    assert_equal(1, s[-3])
    assert_equal(nil, s[4])
    assert_equal(nil, s[-4])

    assert_equal([1, 2], s[0, 2])
    assert_equal([1, 2], s[0..1])
    assert_equal([1, 2], s[-3..-2])
    assert_equal([], s[2..0])
    assert_equal(nil, s[0, -1])

    assert_equal(3, s.pop)
    assert_equal(2, s.pop)
    assert_equal(1, s.pop)
    assert_equal(nil, s.pop)

    s.push(1)
    s.push(2)
    s.push(3)

    assert_equal([1, 2, 3], s.to_a)

    assert_equal(1, s.shift)
    assert_equal(2, s.shift)
    assert_equal(3, s.shift)
    assert_equal(nil, s.pop)

    s.unshift(1)
    s.unshift(2)
    s.unshift(3)

    assert_equal([3, 2, 1], s.to_a)

    assert_equal(1, s.pop)
    assert_equal(2, s.pop)
    assert_equal(3, s.pop)
    assert_equal(nil, s.pop)
  end

  class Counter < Koya::KoyaObject
    def initialize
      self.value = 0
    end
    koya_attr(:value)
    
    def up
      transaction do
        self.value = self.value + 1
      end
    end

    def down
      transaction do
        set_value(value() - 1)
      end
    end
  end
  
  def test_one_db_two_threads
    @koya.transaction do
      @koya.root['count'] = Counter.new
    end
    c = @koya.root['count']
    
    t1 = Thread.new do
      30.times do 
        c.up
      end
    end

    t2 = Thread.new do
      30.times do 
        c.down
      end
    end

    t1.join
    t2.join
    
    assert_equal(0, c.value)
  end

  def test_extent
    bag = @koya.transaction do
      @koya.root['bag'] = Koya::Stream.new
    end

    bag.transaction do
      bag.push MyNode.new
      bag.push MyKoya.new
      bag.push MyCell.new(bag[1])
      bag.push Counter.new
      bag.push Counter.new
      bag.push Counter.new
      bag.push MyNode.new
      bag.push MyCell.new(bag[1])
      bag.push MyCell.new(bag[1])
    end

    assert_equal(9, bag.size)

    ary = @koya.extent(Counter)
    
    assert_equal(3, ary.size)
    assert_kind_of(Counter, ary[0])

    ary = @koya.extent(MyKoya)

    assert_equal(1, ary.size)
    ary[0].name = 'foo'

    assert_equal('foo', bag[1].name)
    
    assert_equal(ary[0], bag[1])

    ary = bag[1]._koya_.referer
    assert_equal(4, ary.size)

    bag[7].value = nil
    ary = bag[1]._koya_.referer
    assert_equal(3, ary.size)

    bag.pop
    ary = bag[1]._koya_.referer
    assert_equal(2, ary.size)

    assert_equal(2, @koya.extent(MyCell).size)

    bag.pop
    assert_equal(1, @koya.extent(MyCell).size)
  end

  def test_revert_to
    bag = @koya.transaction do
      @koya.root['bag'] = Koya::Stream.new
    end

    bag.transaction do
      bag.push MyNode.new
      bag.push MyKoya.new
      bag.push MyCell.new(bag[1])
      bag.push Counter.new
      bag.push Counter.new
      bag.push Counter.new
      bag.push MyNode.new
      bag.push MyCell.new(bag[1])
      bag.push MyCell.new(bag[1])
    end

    rev = Time.now

    assert_equal(9, bag.size)
    
    bag.transaction do
      bag.push MyNode.new
      bag.push MyKoya.new
      bag.push MyCell.new(bag[1])
      bag.push Counter.new
      bag.push Counter.new
      bag.push Counter.new
      bag.push MyNode.new
      bag.push MyCell.new(bag[1])
      bag.push MyCell.new(bag[1])
    end
    bag[7].value = nil

    assert_equal(nil, bag[7].value)
    assert_equal(18, bag.size)

    rev2 = Time.now
    @koya.revert_to(rev)

    assert_equal(9, bag.size)
    assert_equal(bag[1], bag[7].value)

    @koya.revert_to(rev2)

    assert_equal(nil, bag[7].value)
    assert_equal(18, bag.size)
  end
end

class KoyaWOLogTest < KoyaTest
  def setup
    @koya = Koya::Store.new(@dbname, false)
  end

  def test_revert_to
    bag = @koya.transaction do
      @koya.root['bag'] = Koya::Stream.new
    end

    rev = Time.now

    bag.transaction do
      bag.push(1)
    end

    assert_raise(Koya::CanNotRevertError) do
      @koya.revert_to(rev)
    end
  end
end

